"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ix_1 = require("ix");
const jsc = require("jscodeshift");
const js_module_1 = require("./js-module");
const url_converter_1 = require("./url-converter");
const util_1 = require("./util");
class BaseConverter {
    constructor(analysis, options) {
        this.outputs = new Map();
        this.namespacedExports = new Map();
        this._analysis = analysis;
        this.namespaces =
            new Set(util_1.getNamespaces(analysis).concat(options.namespaces || []));
        const referenceRewrites = new Map();
        const windowDotDocument = jsc.memberExpression(jsc.identifier('window'), jsc.identifier('document'));
        referenceRewrites.set('document.currentScript.ownerDocument', windowDotDocument);
        this.referenceRewrites = referenceRewrites;
        const dangerousReferences = new Map();
        dangerousReferences.set('document.currentScript', `document.currentScript is always \`null\` in an ES6 module.`);
        this.dangerousReferences = dangerousReferences;
        this.excludes = new Set(options.excludes);
        this.referenceExcludes = new Set(options.referenceExcludes);
        const importedFiles = ix_1.Iterable
            .from(this._analysis.getFeatures({ kind: 'import', externalPackages: false }))
            .map((imp) => imp.url)
            .filter((url) => !(url.startsWith('bower_components') ||
            url.startsWith('node_modules')));
        this.includes = new Set(importedFiles);
    }
    async convert() {
        const htmlDocuments = [...this._analysis.getFeatures({ kind: 'html-document' })]
            .filter((d) => {
            return !this.excludes.has(d.url) && d.url && this.filter(d);
        });
        const results = new Map();
        for (const document of htmlDocuments) {
            try {
                this.includes.has(document.url) ?
                    this.convertDocument(document, new Set()) :
                    this.convertHtmlToHtml(document, new Set());
            }
            catch (e) {
                console.error(`Error in ${document.url}`, e);
            }
        }
        for (const convertedModule of this.outputs.values()) {
            if (convertedModule.url.endsWith('shadycss/entrypoints/apply-shim.js') ||
                convertedModule.url.endsWith('shadycss/entrypoints/custom-style-interface.js')) {
                // These are already ES6, and messed with in url-converter.
                continue;
            }
            if (convertedModule.type === 'delete-file') {
                results.set(convertedModule.url, undefined);
            }
            else {
                results.set(convertedModule.url, convertedModule.source);
            }
        }
        return results;
    }
    /**
     * Converts a Polymer Analyzer HTML document to a JS module
     */
    convertDocument(document, visited) {
        const jsUrl = url_converter_1.convertHtmlDocumentUrl(url_converter_1.getDocumentUrl(document));
        if (!this.outputs.has(jsUrl)) {
            const newModules = this.getDocumentConverter(document, visited).convertToJsModule();
            this.handleNewJsModules(newModules);
        }
    }
    /**
     * Converts HTML Imports and inline scripts to module scripts as necessary.
     */
    convertHtmlToHtml(document, visited) {
        const htmlUrl = `./${document.url}`;
        if (!this.outputs.has(htmlUrl)) {
            const newModules = this.getDocumentConverter(document, visited)
                .convertAsToplevelHtmlDocument();
            this.handleNewJsModules(newModules);
        }
    }
    handleNewJsModules(outputs) {
        for (const output of outputs) {
            this.outputs.set(output.url, output);
            if (output.type === 'js-module') {
                for (const expr of output.exportedNamespaceMembers) {
                    this.namespacedExports.set(expr.oldNamespacedName, new js_module_1.JsExport(output.url, expr.es6ExportName));
                }
            }
        }
    }
    filter(_document) {
        return true;
    }
}
exports.BaseConverter = BaseConverter;
//# sourceMappingURL=base-converter.js.map