"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("mz/fs");
const path = require("path");
const polymer_analyzer_1 = require("polymer-analyzer");
const rimraf = require("rimraf");
const analysis_converter_1 = require("./analysis-converter");
const manifest_converter_1 = require("./manifest-converter");
const special_casing_1 = require("./special-casing");
const mkdirp = require('mkdirp');
function configureAnalyzer(options) {
    const inDir = options.inDir || process.cwd();
    const urlLoader = new polymer_analyzer_1.InMemoryOverlayUrlLoader(new polymer_analyzer_1.FSUrlLoader(inDir));
    for (const [url, contents] of special_casing_1.polymerFileOverrides) {
        urlLoader.urlContentsMap.set(url, contents);
        urlLoader.urlContentsMap.set(`bower_components/polymer/${url}`, contents);
    }
    return new polymer_analyzer_1.Analyzer({
        urlLoader,
        urlResolver: new polymer_analyzer_1.PackageUrlResolver(),
    });
}
exports.configureAnalyzer = configureAnalyzer;
function configureConverter(analysis, options) {
    return new analysis_converter_1.AnalysisConverter(analysis, {
        packageName: options.packageName,
        packageType: 'element',
        namespaces: options.namespaces,
        excludes: [...(options.excludes || []), 'neon-animation/web-animations.html'],
        referenceExcludes: options.referenceExcludes ||
            [
                'Polymer.Settings',
                'Polymer.log',
                'Polymer.rootPath',
                'Polymer.sanitizeDOMValue',
                'Polymer.Collection',
            ],
        mainFiles: options.mainFiles
    });
}
exports.configureConverter = configureConverter;
/**
 * Converts an entire package from HTML imports to JS modules
 */
async function convertPackage(options) {
    const outDir = options.outDir || 'js_out';
    const outDirResolved = path.resolve(process.cwd(), outDir);
    console.log(`Out directory: ${outDirResolved}`);
    const npmPackageName = options.packageName;
    const npmPackageVersion = options.prereleaseVersion ?
        `${options.packageVersion}-${options.prereleaseVersion}` :
        options.packageVersion;
    const analyzer = configureAnalyzer(options);
    const analysis = await analyzer.analyzePackage();
    const converter = configureConverter(analysis, options);
    const results = await converter.convert();
    if (options.clearOutDir) {
        rimraf.sync(outDirResolved);
    }
    try {
        await fs.mkdir(outDirResolved);
    }
    catch (e) {
        if (e.errno === -17) {
            // do nothing
        }
        else {
            throw e;
        }
    }
    for (const [newPath, newSource] of results) {
        const outPath = path.resolve(outDirResolved, newPath);
        mkdirp.sync(path.dirname(outPath));
        if (newSource !== undefined) {
            await fs.writeFile(outPath, newSource);
        }
        else {
            await fs.unlink(outPath);
        }
    }
    try {
        const bowerJson = manifest_converter_1.readJson(path.join(options.inDir || '.', 'bower.json'));
        const packageJson = manifest_converter_1.generatePackageJson(bowerJson, npmPackageName, npmPackageVersion);
        manifest_converter_1.writeJson(packageJson, outDir, 'package.json');
    }
    catch (e) {
        console.log('error in bower.json -> package.json conversion');
        console.error(e);
    }
}
exports.convertPackage = convertPackage;
//# sourceMappingURL=convert-package.js.map