"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const mz_1 = require("mz");
const path = require("path");
const polymer_analyzer_1 = require("polymer-analyzer");
const semver = require("semver");
const manifest_converter_1 = require("./manifest-converter");
const special_casing_1 = require("./special-casing");
const workspace_converter_1 = require("./workspace-converter");
function configureAnalyzer(options) {
    const inDir = options.inDir || process.cwd();
    const urlLoader = new polymer_analyzer_1.InMemoryOverlayUrlLoader(new polymer_analyzer_1.FSUrlLoader(inDir));
    for (const [url, contents] of special_casing_1.polymerFileOverrides) {
        urlLoader.urlContentsMap.set(`polymer/${url}`, contents);
    }
    return new polymer_analyzer_1.Analyzer({
        urlLoader,
        urlResolver: new polymer_analyzer_1.PackageUrlResolver(),
    });
}
exports.configureAnalyzer = configureAnalyzer;
function configureConverter(analysis, options) {
    return new workspace_converter_1.WorkspaceConverter(analysis, {
        namespaces: options.namespaces,
        excludes: [...(options.excludes || []), 'neon-animation/web-animations.html'],
        referenceExcludes: options.referenceExcludes ||
            [
                'Polymer.DomModule',
                'Polymer.Settings',
                'Polymer.log',
                'Polymer.rootPath',
                'Polymer.sanitizeDOMValue',
                'Polymer.Collection',
            ],
    });
}
exports.configureConverter = configureConverter;
async function convertWorkspace(options) {
    console.log('Converting Workspace');
    const analyzer = configureAnalyzer(options);
    const analysis = await analyzer.analyzePackage();
    const converter = configureConverter(analysis, options);
    const results = await converter.convert();
    for (const [outUrl, newSource] of results) {
        const outPath = path.join(options.inDir, outUrl);
        if (newSource === undefined) {
            await mz_1.fs.unlink(outPath);
        }
        else {
            await mz_1.fs.writeFile(outPath, newSource);
        }
    }
    for (const repo of await mz_1.fs.readdir(options.inDir)) {
        const bowerPath = path.join(options.inDir, repo, 'bower.json');
        const bowerExists = await mz_1.fs.exists(bowerPath);
        if (!bowerExists) {
            continue;
        }
        try {
            const bowerJson = manifest_converter_1.readJson(bowerPath);
            // TODO(https://github.com/Polymer/polymer-modulizer/issues/122):
            // unhardcode
            const bowerName = bowerJson.name;
            let depMapping = manifest_converter_1.dependencyMap[bowerName];
            if (!depMapping) {
                console.warn(`"${bowerName}" npm mapping not found`);
                depMapping = { npm: bowerName };
            }
            let npmPackageVersion = options.packageVersion ||
                (bowerJson.version && semver.inc(bowerJson.version, 'major')) ||
                '3.0.0';
            if (options.prereleaseVersion) {
                npmPackageVersion = `${npmPackageVersion}-${options.prereleaseVersion}`;
            }
            const packageJson = manifest_converter_1.generatePackageJson(bowerJson, depMapping.npm, npmPackageVersion);
            manifest_converter_1.writeJson(packageJson, options.inDir, repo, 'package.json');
        }
        catch (e) {
            console.log('error in bower.json -> package.json conversion');
            console.error(e);
        }
    }
}
exports.convertWorkspace = convertWorkspace;
//# sourceMappingURL=convert-workspace.js.map