"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const astTypes = require("ast-types");
const dom5 = require("dom5");
const ix_1 = require("ix");
const jsc = require("jscodeshift");
const parse5 = require("parse5");
const path = require("path");
function serializeNode(node) {
    const container = parse5.treeAdapters.default.createDocumentFragment();
    dom5.append(container, node);
    return parse5.serialize(container);
}
exports.serializeNode = serializeNode;
/**
 * Get the import name for an imported module object. Useful when generating an
 * import statement, or a reference to an imported module object.
 */
function getModuleId(url) {
    const baseName = path.basename(url);
    const lastDotIndex = baseName.lastIndexOf('.');
    const mainName = baseName.substring(0, lastDotIndex);
    return dashToCamelCase(mainName);
}
exports.getModuleId = getModuleId;
/**
 * Finds an unused identifier name given a requested name and set of used names.
 */
function findAvailableIdentifier(requested, used) {
    let suffix = 0;
    let alias = requested;
    while (used.has(alias)) {
        alias = requested + '$' + (suffix++);
    }
    return alias;
}
exports.findAvailableIdentifier = findAvailableIdentifier;
function dashToCamelCase(s) {
    return s.replace(/-[a-z]/g, (m) => m[1].toUpperCase());
}
function sourceLocationsEqual(a, b) {
    if (a === b) {
        return true;
    }
    const aLoc = a.loc;
    const bLoc = b.loc;
    if (aLoc === bLoc) {
        return true;
    }
    if (aLoc == null || bLoc == null) {
        return false;
    }
    return aLoc.start.column === bLoc.start.column &&
        aLoc.start.line === bLoc.start.line &&
        aLoc.end.column === bLoc.end.column && aLoc.end.line === bLoc.end.line;
}
exports.sourceLocationsEqual = sourceLocationsEqual;
function nodeToTemplateLiteral(node, addNewlines = true) {
    const lines = parse5.serialize(node).split('\n');
    // Remove empty / whitespace-only leading lines.
    while (/^\s*$/.test(lines[0])) {
        lines.shift();
    }
    // Remove empty / whitespace-only trailing lines.
    while (/^\s*$/.test(lines[lines.length - 1])) {
        lines.pop();
    }
    let cooked = lines.join('\n');
    if (addNewlines) {
        cooked = `\n${cooked}\n`;
    }
    // The `\` -> `\\` replacement must occur first so that the backslashes
    // introduced by later replacements are not replaced.
    const raw = cooked.replace(/<\/script/g, '&lt;/script')
        .replace(/\\/g, '\\\\')
        .replace(/`/g, '\\`')
        .replace(/\$/g, '\\$');
    return jsc.templateLiteral([jsc.templateElement({ cooked, raw }, true)], []);
}
exports.nodeToTemplateLiteral = nodeToTemplateLiteral;
/**
 * Returns an array of identifiers if an expression is a chain of property
 * access, as used in namespace-style exports.
 */
function getMemberPath(expression) {
    if (expression.type !== 'MemberExpression' || expression.computed ||
        expression.property.type !== 'Identifier') {
        return;
    }
    const property = expression.property.name;
    if (expression.object.type === 'ThisExpression') {
        return ['this', property];
    }
    else if (expression.object.type === 'Identifier') {
        if (expression.object.name === 'window') {
            return [property];
        }
        else {
            return [expression.object.name, property];
        }
    }
    else if (expression.object.type === 'MemberExpression') {
        const prefixPath = getMemberPath(expression.object);
        if (prefixPath !== undefined) {
            return [...prefixPath, property];
        }
    }
    return undefined;
}
exports.getMemberPath = getMemberPath;
/** Like getMemberPath but joins the name back up into a single string. */
function getMemberName(expression) {
    const path = getMemberPath(expression);
    return path ? path.join('.') : path;
}
exports.getMemberName = getMemberName;
function getMemberOrIdentifierName(expression) {
    if (expression.type === 'Identifier') {
        return expression.name;
    }
    return getMemberName(expression);
}
exports.getMemberOrIdentifierName = getMemberOrIdentifierName;
/**
 * Find the node in program that corresponds to the same part of code
 * as the given node.
 *
 * This method is necessary because program is parsed locally, but the
 * given node may have come from the analyzer (so a different parse run,
 * possibly by a different parser, though they output the same format).
 */
function getNodeGivenAnalyzerAstNode(program, node) {
    const nodePath = getNodePathGivenAnalyzerAstNode(program, node);
    return nodePath ? nodePath.node : undefined;
}
exports.getNodeGivenAnalyzerAstNode = getNodeGivenAnalyzerAstNode;
/** Like `getNode`, but returns the `NodePath` for mutating the AST. */
function getNodePathGivenAnalyzerAstNode(program, node) {
    let associatedNodePath;
    astTypes.visit(program, {
        visitNode(path) {
            // Traverse first, because we want the most specific node that exactly
            // matches the given node.
            this.traverse(path);
            if (associatedNodePath === undefined &&
                sourceLocationsEqual(path.node, node) &&
                path.node.type === node.type) {
                associatedNodePath = path;
                return false;
            }
            return undefined;
        }
    });
    return associatedNodePath;
}
exports.getNodePathGivenAnalyzerAstNode = getNodePathGivenAnalyzerAstNode;
/**
 * Yields the NodePath for each statement at the top level of `program`.
 *
 * Like `yield* program.body` except it yields NodePaths rather than
 * Nodes, so that the caller can mutate the AST with the NodePath api.
 */
function* toplevelStatements(program) {
    const nodePaths = [];
    astTypes.visit(program, {
        visitNode(path) {
            if (!path.parent) {
                // toplevel program
                this.traverse(path);
                return;
            }
            if (path.parent.node.type !== 'Program') {
                // not a toplevel statement, skip it
                return false;
            }
            this.traverse(path);
            // ok, path.node must be a toplevel statement of the program.
            nodePaths.push(path);
            return;
        }
    });
    yield* nodePaths;
}
exports.toplevelStatements = toplevelStatements;
/**
 * Returns true if a statement is the literal "use strict".
 */
function isUseStrict(statement) {
    return statement.type === 'ExpressionStatement' &&
        statement.expression.type === 'Literal' &&
        statement.expression.value === 'use strict';
}
exports.isUseStrict = isUseStrict;
function getNamespaces(analysis) {
    return ix_1.Iterable
        .from(analysis.getFeatures({ kind: 'namespace', externalPackages: true, imported: true }))
        .map((n) => {
        const name = n.name;
        if (name.startsWith('window.')) {
            return name.slice('window.'.length);
        }
        return name;
    });
}
exports.getNamespaces = getNamespaces;
//# sourceMappingURL=util.js.map