"use strict";
/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const pad = require("pad");
const ProgressBar = require("progress");
// TODO(usergenic): import doesn't seem to work for escape-string-regexp.
const escapeStringRegexp = require('escape-string-regexp');
// TODO(usergenic): Replace magic numbers with configurable or
// viewport-dimension-derived values.
const progressMessageWidth = 40;
const progressBarWidth = 45;
/**
 * Synchronously determines whether the given file exists.
 */
function existsSync(fn) {
    return safeStatSync(fn) !== undefined;
}
exports.existsSync = existsSync;
/**
 * Synchronously determines whether the given file exists.
 */
function isDirSync(fn) {
    const stats = safeStatSync(fn);
    if (stats === undefined) {
        return false;
    }
    return stats.isDirectory();
}
exports.isDirSync = isDirSync;
/**
 * Synchronously determines whether the given file exists.
 */
function safeStatSync(fn) {
    try {
        return fs.statSync(fn);
    }
    catch (_) {
        return undefined;
    }
}
exports.safeStatSync = safeStatSync;
/**
 * Like Promise.all, but also displays a progress bar that fills as the
 * promises resolve. The label is a helpful string describing the operation
 * that the user is waiting on.
 */
function promiseAllWithProgress(promises, label) {
    if (promises.length === 0) {
        return Promise.all([]);
    }
    const progressBar = standardProgressBar(label, promises.length);
    const progressed = [];
    for (const promise of promises) {
        let res;
        progressed.push(Promise.resolve(promise)
            .then((resolution) => {
            res = resolution;
            return progressBar.tick();
        })
            .then(() => res));
    }
    return Promise.all(progressed);
}
exports.promiseAllWithProgress = promiseAllWithProgress;
/**
 * Factory function for Progress Bar instance establishing conventional label
 * and progress number format.
 */
function standardProgressBar(label, total) {
    const pb = new ProgressBar(`${pad(label, progressMessageWidth, { strip: true })} [:bar] :percent`, { total, width: progressBarWidth });
    // force the progress bar to start at 0%
    pb.render();
    return pb;
}
exports.standardProgressBar = standardProgressBar;
function progressMessage(msg) {
    return pad(msg, progressMessageWidth, { strip: true });
}
exports.progressMessage = progressMessage;
/**
 * @returns a regular expression where all '*' characters in provided pattern
 * match any character but all other characters in pattern match only string
 * literals.
 */
function wildcardRegExp(pattern) {
    return new RegExp('^' + escapeStringRegexp(pattern).replace(/\\\*/g, '.*') + '$', 'i');
}
exports.wildcardRegExp = wildcardRegExp;
//# sourceMappingURL=util.js.map